"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Neighbors = void 0;
const errors_1 = require("./errors");
/**
 * The representation of all neighbors of a cell
 */
class Neighbors {
    grid;
    coordinate;
    /**
     * @param grid The grid the cell is part of
     * @param coordinate The coordinate in the grid
     */
    constructor(grid, coordinate) {
        this.grid = grid;
        this.coordinate = coordinate;
    }
    /**
     * @param direction The direction to the neighbor cell
     * @returns true if a cell in the given direction exists in the grid
     */
    exists(direction) {
        const offset = this.getOffsetCoordinate(direction);
        const neighbor = {
            row: this.coordinate.row + offset.row,
            col: this.coordinate.col + offset.col,
        };
        return this.grid.cellExists(neighbor);
    }
    /**
     * @param direction The direction to the neighbor cell
     * @returns The coordinate of the neighbor cell
     * @throws {NeighborDoesNotExistInGridError} when the neighbor cell does not exist in the grid
     */
    getCoordinate(direction) {
        if (!this.exists(direction))
            throw new errors_1.NeighborDoesNotExistInGridError(this.grid, this.coordinate, direction);
        const offset = this.getOffsetCoordinate(direction);
        return {
            row: this.coordinate.row + offset.row,
            col: this.coordinate.col + offset.col,
            source: this.coordinate,
            direction,
        };
    }
    /**
     * @param directions The allowed directions
     * @returns An array of all existing neighbor cell coordinates
     */
    listCoordinates(directions) {
        return directions
            .filter((direction) => this.exists(direction))
            .reduce((neighbors, direction) => [...neighbors, this.getCoordinate(direction)], []);
    }
    /**
     * @param direction The direction to the neighbor cell
     * @returns The neighbor cell
     * @throws {NeighborDoesNotExistInGridError} when the neighbor cell does not exist in the grid
     */
    get(direction) {
        if (!this.exists(direction))
            throw new errors_1.NeighborDoesNotExistInGridError(this.grid, this.coordinate, direction);
        const neighbor = this.getCoordinate(direction);
        return neighbor && this.grid.getCell(neighbor);
    }
    /**
     * @param directions The allowed directions
     * @returns An array of all existing neighbor cells
     */
    list(directions) {
        return this.listCoordinates(directions).map((coordinate) => this.grid.getCell(coordinate));
    }
}
exports.Neighbors = Neighbors;
