import { Row } from './Row';
import { Column } from './Column';
import { CellDoesNotExistInGridError, InvalidGridSizeError, UnequalGridHeightError, UnequalGridWidthError, } from './errors';
import { GridEventDispatcher } from './utils';
/**
 * The grid contains all information about cells
 */
export class Grid {
    options;
    width;
    height;
    _grid = [];
    _rows = [];
    _columns = [];
    eventDispatcher = new GridEventDispatcher();
    /**
     * @param options The initialization configuration
     * @param initializeCell This function will initialize a cell
     */
    constructor(options) {
        this.options = options;
        if ('initializeCellValue' in this.options) {
            this.width = this.options.width;
            this.height = this.options.height;
        }
        else if ('grid' in this.options) {
            const rows = this.options.grid;
            this.height = rows.length;
            this.width = Math.max(0, ...rows.map((row) => row.length));
        }
        else {
            throw `Invalid options provided`;
        }
    }
    initialize(initializeCell) {
        if ('initializeCellValue' in this.options) {
            this._grid.push(...this.initializeCells(initializeCell, this.options.initializeCellValue));
        }
        if ('grid' in this.options) {
            const rows = this.options.grid;
            this._grid.push(...this.initializeCells(initializeCell, ({ row, col }) => rows[row][col]));
        }
        for (let row = 0; row < this.height; row++) {
            this._rows.push(new Row(this, row));
        }
        for (let col = 0; col < this.width; col++) {
            this._columns.push(new Column(this, col));
        }
        const forwardEvent = (event) => this.eventDispatcher.dispatchCellValueChangedEvent(event);
        for (const cell of this.cells) {
            cell.onValueChanged(forwardEvent);
        }
    }
    initializeCells(initializeCell, initializeCellValue) {
        if (this.height === 0 || this.width === 0)
            throw new InvalidGridSizeError(this.width, this.height);
        const cells = new Array(this.height);
        for (let row = 0; row < this.height; row++) {
            cells[row] = new Array(this.width);
            for (let col = 0; col < this.width; col++) {
                const coordinate = { row, col };
                const value = initializeCellValue(coordinate);
                cells[row][col] = initializeCell(coordinate, value);
            }
        }
        return cells;
    }
    /**
     * @returns The 2-dimensional array representing a grid (1. dimension: row (y-axis), 2. dimension: column (x-axis))
     */
    get grid() {
        return this._grid;
    }
    /**
     * @param coordinate The coordinate of the cell
     * @returns true if the cell exists in the grid
     */
    cellExists(coordinate) {
        const { row, col } = coordinate;
        return col >= 0 && row >= 0 && row < this.height && col < this.width;
    }
    /**
     * @returns All cells in the grid in ascending order by row and column
     */
    get cells() {
        return this.grid.flat();
    }
    /**
     * @param coordinate The coordinate of the cell
     * @returns The cell
     * @throws {CellDoesNotExistInGridError} when the cell does not exist in the grid
     */
    getCell(coordinate) {
        if (!this.cellExists(coordinate))
            throw new CellDoesNotExistInGridError(this, coordinate);
        return this.grid[coordinate.row][coordinate.col];
    }
    /**
     * @param row The row coordinate
     * @returns The row
     */
    getRow(row) {
        return this._rows[row];
    }
    /**
     * @returns All rows of the grid in ascending order
     */
    get rows() {
        return this._rows;
    }
    /**
     * @param col The column coordinate
     * @returns The column
     */
    getColumn(col) {
        return this._columns[col];
    }
    /**
     * @returns All columns of the grid in ascending order
     */
    get columns() {
        return this._columns;
    }
    /**
     * @param callback A function that should be called, when a cell value of this grid changes
     * @returns a function to unregister the callback
     */
    onCellValueChanged(callback) {
        return this.eventDispatcher.onCellValueChanged(callback);
    }
    /**
     * @param gridToAdd The grid that should be added to this grid at the given direction
     * @param addDirection The direction at that the given grid should be placed
     * @returns a new grid
     */
    extend(gridToAdd, addDirection) {
        if ((addDirection === 'LEFT' || addDirection === 'RIGHT') && this.height !== gridToAdd.height)
            throw new UnequalGridHeightError(this.height, gridToAdd.height);
        if ((addDirection === 'TOP' || addDirection === 'BOTTOM') && this.width !== gridToAdd.width)
            throw new UnequalGridWidthError(this.width, gridToAdd.width);
        const extractValues = (row) => row.cells.map((cell) => cell.value);
        switch (addDirection) {
            case 'TOP':
                return this.initializeGrid({
                    grid: [...gridToAdd.rows.map(extractValues), ...this.rows.map(extractValues)],
                });
            case 'BOTTOM':
                return this.initializeGrid({
                    grid: [...this.rows.map(extractValues), ...gridToAdd.rows.map(extractValues)],
                });
            case 'LEFT':
                return this.initializeGrid({
                    grid: this.rows.map((row) => [...extractValues(gridToAdd.getRow(row.row)), ...extractValues(row)]),
                });
            case 'RIGHT':
                return this.initializeGrid({
                    grid: this.rows.map((row) => [...extractValues(row), ...extractValues(gridToAdd.getRow(row.row))]),
                });
        }
    }
    /**
     * @param topLeft The coordinate of the top/left where the new grid should begin
     * @param bottomRight The coordinate of the bottom/right where the new grid should end
     * @returns a new grid
     */
    crop(topLeft, bottomRight) {
        const width = bottomRight.col - topLeft.col;
        const height = bottomRight.row - topLeft.row;
        if (width <= 0 || height <= 0)
            throw new InvalidGridSizeError(width, height);
        return this.initializeGrid({
            grid: this.rows
                .filter(({ row }) => row >= topLeft.row)
                .filter(({ row }) => row <= bottomRight.row)
                .map((row) => row.cells.filter(({ col }) => col >= topLeft.col).filter(({ col }) => col <= bottomRight.col))
                .map((cells) => cells.map((cell) => cell.value)),
        });
    }
    /**
     * @param cloneValue A custom function to clone the value of a cell (defaults to copying the value)
     * @returns The cloned grid
     */
    clone(cloneValue = (value) => value) {
        return this.initializeGrid({
            width: this.width,
            height: this.height,
            initializeCellValue: ({ row, col }) => cloneValue(this.grid[row][col].value),
        });
    }
}
